/* Copyright (C) Gabriel T. Delarosa, 2010. 
 * All rights reserved worldwide.
 *
 * This software is provided "as is" without express or implied
 * warranties. You may freely copy and compile this source into
 * applications you distribute provided that the authors name is
 * mentioned.
 */

#include "RendererBase.h"

namespace vi
{

RendererBase::RendererBase() :
    d3d_( NULL ),
    d3dDevice_( NULL )
{
};

RendererBase::~RendererBase()
{
    if( d3dDevice_ != NULL )
        d3dDevice_->Release();

    if( d3d_ != NULL )
        d3d_->Release();
}

HRESULT RendererBase::initD3D( HWND hWnd )
{
    if( ( d3d_ = Direct3DCreate9( D3D_SDK_VERSION ) )==NULL )
        return E_FAIL;

    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

    D3DCAPS9 Caps;
    DWORD dBehaviour;
    d3d_->GetDeviceCaps(D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,&Caps);
    if( Caps.VertexShaderVersion < D3DVS_VERSION(1,1) )
    {
        dBehaviour = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
        MessageBox(hWnd,"Using software vertex processing...\nPerformance will be slow",
            "Warning",MB_OK);

    }
    else
        dBehaviour = D3DCREATE_HARDWARE_VERTEXPROCESSING;

    if( FAILED( d3d_->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                dBehaviour,
                                &d3dpp, &d3dDevice_ 
                            ) 
                )
        )
    {
        return E_FAIL;
    }

    d3dDevice_->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

    d3dDevice_->SetRenderState( D3DRS_LIGHTING, FALSE );

    //d3dDevice_->SetRenderState( D3DRS_ZENABLE, TRUE );

    return S_OK;
}

void RendererBase::render( QCamera &cam, const ViewPort &vp, const Airplane &airplane )
{
    cam.yawPitchRoll( airplane.yaw_, airplane.pitch_, airplane.roll_ );
    cam.translate( 0, airplane.climbRate_, airplane.speed_ );

    //keep camera above ground
    if ( cam.eye().y <= 5.0f )
        cam.translate( 0, 5.0f - cam.eye().y, 0 );

    cam.update();

    // Clear the backbuffer to a blue color
    d3dDevice_->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,
                        D3DCOLOR_XRGB(0,0,125), 1.0f, 0 );


    d3dDevice_->BeginScene();

    D3DXVECTOR3 lookAt = cam.eye() + cam.lookAt();
    D3DXMatrixLookAtRH( &viewMatrix_, &cam.eye(), &lookAt, &cam.up() );

    drawHorizon( cam, vp );

    d3dDevice_->EndScene();

    d3dDevice_->Present( NULL, NULL, NULL, NULL );
}

void RendererBase::drawHorizon( QCamera &qcamera, const ViewPort &viewport )
{
    D3DXVECTOR3 lookAt = qcamera.lookAt();
    D3DXVECTOR3 side = qcamera.side();
    D3DXVECTOR3 up = qcamera.up();
    D3DXVECTOR3 eye = qcamera.eye();
    
    std::vector<D3DXVECTOR3> clippedPts, pts;
    pts.reserve(4);

    D3DXVECTOR3 viewC = lookAt * viewport.viewDistance();
    float scrHalfW = viewport.screenWidth() * 0.5f;
    float scrHalfH = viewport.screenHeight() * 0.5f;

    // UL
    pts.push_back((up * scrHalfH) + side * (-scrHalfW) + viewC); 
    // UR
    pts.push_back((up * scrHalfH) + (side * scrHalfW) + viewC);
    //LR  
    pts.push_back((up * -(scrHalfH)) + (side * scrHalfW) + viewC);
    //LL  
    pts.push_back((up * -(scrHalfH)) + (side * (-scrHalfW)) + viewC);

    int v1,v2, cp = 0;
    float x1,y1,z1,x2,y2,z2,newx,newz,m;
    float ymax = -1.0f;
    v1=3;
    for (v2=0; v2 < 4; v2++)
    {
        x1 = pts[v1].x;
        y1 = pts[v1].y;
        z1 = pts[v1].z;
        x2 = pts[v2].x;
        y2 = pts[v2].y;
        z2 = pts[v2].z;
        if ((y1 <= ymax) && (y2 <= ymax))
        {
            clippedPts.push_back( pts[v2] );
        }
        else if ((y1 > ymax) && (y2 > ymax))
        {
            // completely above
        }
        if ((y1 <= ymax) && (y2 > ymax))
        {
            if (x1 != x2)        
            {                    
                m= (y2-y1) / (x2-x1);                    
                newx = x1 + ((ymax - y1) / m);
            }
            else            
                newx = x1;            
            if (z1 != z2)
            {
                m= (y2-y1) / (z2-z1);                    
                newz = z1 + ((ymax - y1) / m);
            }
            else
                newz = z1;
            clippedPts.push_back( D3DXVECTOR3(newx,ymax,newz) );
        }
        if ((y1 > ymax) && (y2 <= ymax)) 
        {
            if (x1 != x2)        
            {                    
                m= (y2-y1) / (x2-x1);                    
                newx = x1 + ((ymax - y1) / m);
            }
            else            
                newx = x1;            
            if (z1 != z2)
            {
                m= (y2-y1) / (z2-z1);                    
                newz = z1 + ((ymax - y1) / m);
            }
            else
                newz = z1;
            clippedPts.push_back( D3DXVECTOR3(newx,ymax,newz) );
            clippedPts.push_back( pts[v2] );
        }
        v1=v2;
    }
    cp = clippedPts.size();
    if (cp == 0)
    {
        return;
    }

    renderCameraFootprint( clippedPts, eye, viewport.viewDistance() );
}


}// end namespace
