/* Copyright (C) Gabriel T. Delarosa, 2010. 
 * All rights reserved worldwide.
 *
 * This software is provided "as is" without express or implied
 * warranties. You may freely copy and compile this source into
 * applications you distribute provided that the authors name is
 * mentioned.
 */

#include "ScanConverter.h"

namespace vi
{
Edge *ScanConverter::makeEdge( const D3DXVECTOR2& u, const D3DXVECTOR2& v ) const
{
    if ( u.y == v.y )
        return 0;
    else
        return ( u.y > v.y ) ? new Edge( u,v ) : new Edge( v,u );
}

void ScanConverter::insertEdge( const D3DXVECTOR2& u, const D3DXVECTOR2& v )
{
    Edge *newEdge = makeEdge( u, v );

    if ( !newEdge )
        return;

    EdgePtrListIterator cur = edges_.begin();
    while( cur != edges_.end() && (*cur)->yMax_ > newEdge->yMax_ )
    {
        cur++;
    }
    edges_.insert( cur, newEdge );

    //cur = edges_.begin();
    //while( cur != edges_.end() )
    //{
    //    tmp = *cur;
    //    cur++;
    //}

    scan_ = (*edges_.begin())->yMax_ + 1;
    rightEdge_ = edges_.begin();
}

void ScanConverter::adjustRightEdge()
{
    while( rightEdge_ != edges_.end() && (*rightEdge_)->yMax_ >= scan_ )
        rightEdge_++;
}

void ScanConverter::removeScannedEdges()
{
    EdgePtrListIterator cur = edges_.begin();

    while( cur != rightEdge_ )
    {
        if ( (*cur)->yMin_ >= scan_ )
        {
            EdgePtrListIterator tmp = cur;
            Edge *edge = (*tmp);
            tmp++;
            edges_.erase( cur );
            delete edge;
            cur = tmp;
        }
        else
            cur++;
    }
}

inline void ScanConverter::updateEdge( const EdgePtrListIterator &eIt )
{
    Edge *edge = (*eIt);
    edge->update();
}
inline void ScanConverter::getFrontTwoEdges( EdgePtrListIterator &head, EdgePtrListIterator &next )
{
    head = edges_.begin();
    next = head;
    next++;
}

void ScanConverter::rasterize( CellVector &outputCells )
{
    float startY = scan_;
    if ( edges_.begin() == edges_.end() )
        return;

    EdgePtrListIterator head, next;
    int x, rightX, leftX;

    do
    {
        scan_ -= deltaScan_;

        adjustRightEdge();
        removeScannedEdges();

        if ( edges_.begin() == edges_.end() )
            break;
        
        getFrontTwoEdges( head, next );

        updateEdge( head );
        
        if ( next != edges_.end() )
        {
            updateEdge( next );

            //swap edges so the one with least X is on left
            if ( (*head)->curX_ > (*next)->curX_ || 
                  ((*head)->curX_ == (*next)->curX_ &&
                   (*head)->curX_+(*head)->deltaX_ > (*next)->curX_+(*next)->deltaX_
                  )
               )
            {
                edges_.splice( head, edges_, next );
                edges_.splice( rightEdge_, edges_, head );
                getFrontTwoEdges( head, next );
            }

            //scan from left edge to right edge
            CellType cell;
            rightX = static_cast<int>( (*next)->curX_ + (*next)->getStep() ); //we add extra cells to the right in order to ensure there are no gaps in the terrain
            leftX = static_cast<int>((*head)->curX_ - max((*head)->getStep(), 1.0f) );//also add extra cells to the left
            x = leftX;
            for ( ; leftX <= rightX; leftX++ )
            {
                cell.first = leftX;
                cell.second = static_cast<int>(scan_);
                outputCells.push_back( cell );
                if ( scan_+deltaScan_ == startY )
                {
                    cell.second = static_cast<int>(scan_+1);
                    outputCells.push_back( cell );
                }
            }
        }
    }while( edges_.begin() != edges_.end() );

    //we add another row of cells at the bottom of the polygon, to ensure no gaps in the terrain
    if ( scan_ != startY )
    {
        CellType cell;
        for ( leftX = x; leftX <= rightX; leftX++ )
        {
            cell.first = leftX;
            cell.second = static_cast<int>(scan_);
            outputCells.push_back( cell );
        }
    }
}

} //end namespace

