/* Copyright (C) Gabriel T. Delarosa, 2010. 
 * All rights reserved worldwide.
 *
 * This software is provided "as is" without express or implied
 * warranties. You may freely copy and compile this source into
 * applications you distribute provided that the authors name is
 * mentioned.
 */

#include "TerrainRenderer.h"
#include "ScanConverter.h"

namespace vi
{
TerrainRenderer::~TerrainRenderer() 
{ 
    if (terrainVerts_) 
        delete terrainVerts_; 
}


HRESULT TerrainRenderer::init( const ViewPort &vp )
{
    if ( PerspectiveRenderer::init( vp ) == S_OK )
    {
        FILE *f = fopen( "Data/Land_H.raw","rb" );
        fread( htmap_, sizeof(htmap_), 1, f);
        fclose(f);
    }

    const int numVerts = (HMAP_DIM+2) * (HMAP_DIM>>1);
    terrainVerts_ = new TERRAINVERTEX[ numVerts ];

    int i = 0;
    float cellSize = MIN_CELL_SIZE;

    float scale = cellSize/(HMAP_DIM);

    for ( int tz = 0; tz < HMAP_DIM-1; tz += 2)
    {
        for ( int tx = 0; tx < HMAP_DIM+1; tx += 2)
        {
            float h = tx == HMAP_DIM ? 0 : getHeight( tx,tz );
            terrainVerts_[i].position = D3DXVECTOR3( tx * scale, h, tz * scale );
            terrainVerts_[i].tu = (terrainVerts_[i].position.x*4.0f) / cellSize;
            terrainVerts_[i].tv = (terrainVerts_[i].position.z*4.0f) / cellSize;
            i++;

            h = tz == HMAP_DIM-2 ? 0 : getHeight( tx,tz+2 );
            terrainVerts_[i].position = D3DXVECTOR3( tx * scale, h, (tz+2) * scale);
            terrainVerts_[i].tu = (terrainVerts_[i].position.x*4.0f) / cellSize;
            terrainVerts_[i].tv = (terrainVerts_[i].position.z*4.0f) / cellSize;
            i++;
        }
    }

    return S_OK;
}

inline float TerrainRenderer::getScale( int iY )
{
    float fy = max( 1.0f, static_cast<float>(iY)/10.0f );
    float dec = fy - (iY/10);
    if ( dec < .7f )
        fy -= dec;
    return min( fy, 12.0f );
}

void TerrainRenderer::renderCameraFootprint( const std::vector<D3DXVECTOR3> &clippedPts, D3DXVECTOR3 eye, float viewDist )
{
    unsigned int cp = static_cast< unsigned int > ( clippedPts.size() );
    D3DXVECTOR3 footPrintVertex;
    float scale;

    std::vector<D3DXVECTOR2> vertices;
    D3DXVECTOR2 newVertex;
    ScanConverter scanner;
    float cellSize = MIN_CELL_SIZE;

    float hScale = getScale( eye.y );

    cellSize *= hScale;

    #define toCellCoord( x ) static_cast<float>(static_cast<int>(x / cellSize))

    unsigned int i=0;
    for (; i < cp; i++)
    {
        D3DXVec3Normalize( &footPrintVertex, &clippedPts[i] );

        scale = fabs(footPrintVertex.y == 0 ? 0 : eye.y / footPrintVertex.y);
        footPrintVertex *= scale;
        footPrintVertex.x += eye.x;
        footPrintVertex.y += eye.y;
        footPrintVertex.z += eye.z;
    
        //convert world vertex to grid cell coorinate
        newVertex.x = footPrintVertex.x < 0 ? footPrintVertex.x - cellSize : footPrintVertex.x;
        newVertex.y = footPrintVertex.z < 0 ? footPrintVertex.z - cellSize : footPrintVertex.z;

        vertices.push_back( newVertex );
        //for every two grid cell coords, add an edge to scan converter
        if ( i > 0 )
        {
            D3DXVECTOR2 u( toCellCoord( vertices[i-1].x ), toCellCoord( vertices[i-1].y ) );
            D3DXVECTOR2 v( toCellCoord( vertices[i].x ), toCellCoord( vertices[i].y ) );
            scanner.insertEdge( u,v );
        }
    }

    //add last edge to scanner with first and last vertex
    D3DXVECTOR2 u( toCellCoord( vertices[i-1].x ), toCellCoord( vertices[i-1].y ) );
    D3DXVECTOR2 v( toCellCoord( vertices[0].x ), toCellCoord( vertices[0].y ) );
    if ( u != v )
    {
        scanner.insertEdge( u,v );
    }

    //scan convert to get grid cells
    ScanConverter::CellVector outputCells;
    scanner.rasterize( outputCells );

    //special case: if looking straight down, draw four cells around camera
    int eyeX = static_cast<int>( toCellCoord( eye.x ) );
    int eyeY = static_cast<int>( toCellCoord( eye.z ) );
    if ( !scanner.hasEdges() && outputCells.empty() )
    {
        ScanConverter::CellType cell;
        cell.first = eyeX;
        cell.second = eyeY;
        outputCells.push_back( cell );
        cell.first = eyeX-1;
        cell.second = eyeY;
        outputCells.push_back( cell );
        cell.first = eyeX-1;
        cell.second = eyeY-1;
        outputCells.push_back( cell );
        cell.first = eyeX;
        cell.second = eyeY-1;
        outputCells.push_back( cell );
    }

    d3dDevice_->SetTransform( D3DTS_VIEW, &viewMatrix_ );
    d3dDevice_->SetTexture(0, textures_[STONE]);
    d3dDevice_->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
    d3dDevice_->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
    d3dDevice_->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE );
    d3dDevice_->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
    d3dDevice_->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
    d3dDevice_->SetSamplerState( 0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );
    d3dDevice_->SetSamplerState( 0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP );
    d3dDevice_->SetSamplerState( 0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP );
    d3dDevice_->SetFVF( D3DFVF_TERRAINVERTEX );

    //go through every grid cell found by scanner, and render
    ScanConverter::CellVector::const_iterator it = outputCells.begin();
    for(; it != outputCells.end(); it++)
    {
        int x = (*it).first;
        int y = (*it).second;
        renderTerrainCell( x,y, eye.y, cellSize );
    }
}

//void TerrainRenderer::renderCell( TERRAINVERTEX *geometryVertices, int x, int y, float camY, float cellSize )
//{
//    geometryVertices[0].position = D3DXVECTOR3( x*cellSize, 0, y*cellSize );
//    geometryVertices[0].tu = 0.0f;
//    geometryVertices[0].tv = 0.0f;
//    geometryVertices[1].position = D3DXVECTOR3( (x+1)*cellSize, 0, y*cellSize );
//    geometryVertices[1].tu = 0.0f;
//    geometryVertices[1].tv = 0.0f;
//    geometryVertices[2].position = D3DXVECTOR3( (x+1)*cellSize, 0, (y+1)*cellSize );
//    geometryVertices[2].tu = 0.0f;
//    geometryVertices[2].tv = 0.0f;
//    geometryVertices[3].position = D3DXVECTOR3( x*cellSize, 0, (y+1)*cellSize );
//    geometryVertices[3].tu = 0.0f;
//    geometryVertices[3].tv = 0.0f;
//    if ( (x % 2 == 0 && y % 2 != 0) || (x % 2 != 0 && y % 2 == 0) )
//    {
//        geometryVertices[0].tu = 0.0f;
//        geometryVertices[0].tv = 0.0f;
//        geometryVertices[1].tu = 1.0f;
//        geometryVertices[1].tv = 0.0f;
//        geometryVertices[2].tu = 1.0f;
//        geometryVertices[2].tv = 1.0f;
//        geometryVertices[3].tu = 0.0f;
//        geometryVertices[3].tv = 1.0f;
//    }
//    d3dDevice_->DrawPrimitiveUP( D3DPT_TRIANGLEFAN,  2, geometryVertices, sizeof(TERRAINVERTEX) );
//}

void TerrainRenderer::renderTerrainCell( int x, int y, float camY, float cellSize )
{
    const int numVerts = (HMAP_DIM+2) * (HMAP_DIM>>1);
    TERRAINVERTEX terrainVerts[ numVerts ];

    int i = 0;

    float tfx = x*cellSize;
    float tfz = y*cellSize;
    
    float hScale = getScale( camY );
    if ( hScale == 1.0f )
    {
        for (i=0; i<numVerts; i++)
        {
            terrainVerts[i] = terrainVerts_[i];
            terrainVerts[i].position.x += tfx;   terrainVerts[i].position.z += tfz;
        }
    }
    else
    {
        cellSize /= hScale;
        int hmapCellSize = static_cast<int>(hScale*HMAP_DIM);
        int bx = (x*hmapCellSize) % HMAP_DIM;
        int bz = (y*hmapCellSize) % HMAP_DIM;
        if ( x < 0 ) bx = HMAP_DIM + bx;
        if ( y < 0 ) bz = HMAP_DIM + bz;
        for ( int tz = 0; tz < HMAP_DIM-1; tz += 2)
        {
            for ( int tx = 0; tx < HMAP_DIM+1; tx += 2)
            {
                terrainVerts[i] = terrainVerts_[i];
                terrainVerts[i].position *= hScale;
                int tu = static_cast<int>((tx*hScale)+bx) % HMAP_DIM;
                int tv = static_cast<int>((tz*hScale)+bz) % HMAP_DIM;
                terrainVerts[i].position.y = getHeight( tu, tv );
                terrainVerts[i].position.x += tfx;   terrainVerts[i].position.z += tfz;
                terrainVerts[i].tu = (terrainVerts[i].position.x*4.0f) / cellSize;
                terrainVerts[i].tv = (terrainVerts[i].position.z*4.0f) / cellSize;
                i++;

                terrainVerts[i] = terrainVerts_[i];
                terrainVerts[i].position *= hScale;
                tu = static_cast<int>((tx*hScale)+bx) % HMAP_DIM;
                tv = static_cast<int>(((tz+2)*hScale)+bz) % HMAP_DIM;
                terrainVerts[i].position.y = getHeight( tu, tv );
                terrainVerts[i].position.x += tfx;   terrainVerts[i].position.z += tfz;
                terrainVerts[i].tu = (terrainVerts[i].position.x*4.0f) / cellSize;
                terrainVerts[i].tv = (terrainVerts[i].position.z*4.0f) / cellSize;
                i++;
            }
        }
    }

    d3dDevice_->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, numVerts-2, terrainVerts, sizeof(TERRAINVERTEX) );
}

}// end namespace
